-- **********************************************************
-- *			pdp1 Microprocessor CPU Core				*
-- *========================================================*
-- * FPGA Project:		16-Bit CPU in Altera SOPC Builder	*
-- * File:				pdp1.vhd							*
-- * Author:			C.H.Ting							*
-- * Description:		pdp1 CPU Block						*
-- *														*
-- * Revision History:										*
-- * Date		By Who		Modification					*
-- * 06/06/05	C.H. Ting	Convert EP24 to 32-bits.		*
-- * 06/10/05	Robyn King	Made compatible with Altera SOPC*
-- *							Builder.					*
-- * 06/27/05	C.H. Ting	Removed Line Drawing Engine.	*
-- * 07/27/05	Robyn King	Cleaned up code.				*
-- * 08/07/10	C.H. Ting	Return to eP32p					*
-- * 11/18/10	C.H. Ting	Port to LatticeXP2 Brevia Kit	*
-- * 02/29/12	 Chen-Hanson Ting Back to pdp1		    	*
-- * 05/18/16   C. H. Ting  Port to PDP1                    *
-- **********************************************************
 
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;


entity pdp1 is 
	generic(width: integer := 18);
	port(
		-- input port
		clr:		in	std_logic;
		clk:		in	std_logic;
		data_i:		in	std_logic_vector(0 to width-1);
		io_read:	out	std_logic;
		io_write:	out	std_logic;
		mem_read:	out	std_logic;
		mem_write:	out	std_logic;
		addr_o:		out	std_logic_vector(0 to 11);
		data_o:		out	std_logic_vector(0 to width-1)
  	);
end entity pdp1;

architecture behavioral of pdp1 is
	signal sum,dif: std_logic_vector(0 to width);
	signal code,condition: std_logic_vector(0 to 4);
	signal flags: std_logic_vector(0 to 5);
	signal phase,next_phase: integer range 0 to 10;
	signal ac,i,io,dis_in: std_logic_vector(0 to width-1);
	signal ac_zero,io_zero: std_logic;
	signal a,p,p_in: std_logic_vector(0 to 11);
	signal ac_in,io_in: std_logic_vector(0 to width-1);
	signal ac_sel: std_logic_vector(0 to 4);
	signal io_sel: std_logic_vector(0 to 3);
	signal flag_sel: std_logic_vector(0 to 2);
	signal data_sel,p_sel: std_logic_vector(0 to 1);
	signal acload,ovload,ioload,pload,iload,phaseload: std_logic;
	signal ov,ov_in,ov_sel,addr_sel,aload: std_logic;
	
-- machine instructions selected by code
--	constant * : std_logic_vector(0 to 4) :="00000";
	constant andd : 	std_logic_vector(0 to 4) :="00001";
	constant iorr  : 	std_logic_vector(0 to 4) :="00010";
	constant xorr  : 	std_logic_vector(0 to 4) :="00011";
	constant xct : 		std_logic_vector(0 to 4) :="00100";
	constant ddx: 		std_logic_vector(0 to 4) :="00101";
	constant dsp  : 	std_logic_vector(0 to 4) :="00110";
	constant cal : 		std_logic_vector(0 to 4) :="00111";
--	constant jda : 		std_logic_vector(0 to 5) :="001111";
	constant lac : 		std_logic_vector(0 to 4) :="01000";
	constant lio : 		std_logic_vector(0 to 4) :="01001";
	constant dac : 		std_logic_vector(0 to 4) :="01010";
	constant dap : 		std_logic_vector(0 to 4) :="01011";
	constant dip : 		std_logic_vector(0 to 4) :="01100";
	constant dio : 		std_logic_vector(0 to 4) :="01101";
	constant dzm : 		std_logic_vector(0 to 4) :="01110";
--	constant *  : std_logic_vector(0 to 4) :="01111";
	constant add : 		std_logic_vector(0 to 4) :="10000";
	constant sub : 		std_logic_vector(0 to 4) :="10001";
	constant idx : 		std_logic_vector(0 to 4) :="10010";
	constant isp : 		std_logic_vector(0 to 4) :="10011";
	constant sad : 		std_logic_vector(0 to 4) :="10100";
	constant sas : 		std_logic_vector(0 to 4) :="10101";
	constant mus : 		std_logic_vector(0 to 4) :="10110";
	constant dis : 		std_logic_vector(0 to 4) :="10111";
	constant jmp : 		std_logic_vector(0 to 4) :="11000";
	constant jsp : 		std_logic_vector(0 to 4) :="11001";
	constant skp : 		std_logic_vector(0 to 4) :="11010";
	constant sft : 		std_logic_vector(0 to 4) :="11011";
	constant law :		std_logic_vector(0 to 4) :="11100";
	constant iot : 		std_logic_vector(0 to 4) :="11101";
--	constant * : 		std_logic_vector(0 to 4) :="11110";
	constant opr : 		std_logic_vector(0 to 4) :="11111";
-- mux to ac register, selected by ac_sel
	constant ior_ac: 	std_logic_vector :="00000";
	constant xor_ac: 	std_logic_vector :="00001";
	constant and_ac: 	std_logic_vector :="00010";
	constant sum_ac: 	std_logic_vector :="00011";
	constant dif_ac: 	std_logic_vector :="00100";
	constant rar_ac: 	std_logic_vector :="00101";
	constant sar_ac: 	std_logic_vector :="00110";
	constant ral_ac: 	std_logic_vector :="00111";
	constant sal_ac: 	std_logic_vector :="01000";
	constant rcr_ac: 	std_logic_vector :="01001";
	constant scr_ac: 	std_logic_vector :="01010";
	constant rcl_ac: 	std_logic_vector :="01011";
	constant dis_ac: 	std_logic_vector :="01100";
	constant mus_ac: 	std_logic_vector :="01101";
	constant law_ac: 	std_logic_vector :="01110";
	constant p_ac: 		std_logic_vector :="01111";
	constant zero_ac: 	std_logic_vector :="10000";
--	constant io_ac: 	std_logic_vector :="10001";
	constant idx_ac: 	std_logic_vector :="10010";
	constant data_ac: 	std_logic_vector :="10011";
	constant cma_ac: 	std_logic_vector :="10100";
	constant addr_ac: 	std_logic_vector :="10101";
	constant code_ac: 	std_logic_vector :="10110";
	constant ddx_ac: 	std_logic_vector :="10111";
-- mux to iio register, selected by io_sel
--	constant ac_io: 	std_logic_vector :="0000";
	constant mus_io: 	std_logic_vector :="0001";
	constant scr_io: 	std_logic_vector :="0010";
	constant rcl_io: 	std_logic_vector :="0011";
	constant rir_io: 	std_logic_vector :="0100";
	constant ril_io: 	std_logic_vector :="0101";
	constant sir_io: 	std_logic_vector :="0110";
	constant sil_io: 	std_logic_vector :="0111";
	constant dis_io: 	std_logic_vector :="1000";
	constant cli_io: 	std_logic_vector :="1001";
	constant data_io: 	std_logic_vector :="1010";
-- mux to p register, selected by p_sel
	constant p1_p: 		std_logic_vector :="00";
	constant p2_p:		std_logic_vector :="01";
	constant a_p:		std_logic_vector :="10";
	constant data_p: 	std_logic_vector :="11";
-- mux to address bus, selected by addr_sel
	constant p_addr: 	std_logic :='0';
	constant a_addr: 	std_logic :='1';
-- mux to ov, selected by ov_sel
	constant sum_ov: 	std_logic :='0';
	constant dif_ov: 	std_logic :='1';
-- mux to data_o bus, selected by data_sel
	constant ac_data: 	std_logic_vector :="00";
	constant io_data: 	std_logic_vector :="01";
	constant alu_data: 	std_logic_vector :="10";
	constant zero_data: std_logic_vector :="11";
-- shift/rotate
	constant ral: 	std_logic_vector :="0001";
	constant ril: 	std_logic_vector :="0010";
	constant rcl: 	std_logic_vector :="0011";
	constant sal: 	std_logic_vector :="0101";
	constant sil: 	std_logic_vector :="0110";
	constant scl: 	std_logic_vector :="0111";
	constant rar: 	std_logic_vector :="1001";
	constant rir: 	std_logic_vector :="1010";
	constant rcr: 	std_logic_vector :="1011";
	constant sar: 	std_logic_vector :="1101";
	constant sir: 	std_logic_vector :="1110";
	constant scr: 	std_logic_vector :="1111";
-- skip	
	constant sza: 	std_logic_vector :="00001";
	constant spa: 	std_logic_vector :="00010";
	constant sma: 	std_logic_vector :="00100";
	constant szo: 	std_logic_vector :="01000";
	constant spi: 	std_logic_vector :="10000";

begin

	with phase select
	code <= data_i(0 to 4) when 0,
		i(0 to 4) when others;
		
	with data_sel select
	data_o <= io(0 to width-1) when io_data,
		ac_in when alu_data,
		(others => '0') when zero_data,
		ac(0 to width-1) when others;
		
	with p_sel select
	p_in <= data_i(6 to width-1) when data_p,
		p + 2 when p2_p,
		a when a_p,
		p + 1 when others;

	sum(0 to width) <=(('0'&ac(0 to width-1)) +
		('0'&data_i(0 to width-1)));
	dif(0 to width) <=(('0'&ac(0 to width-1)) -
		('0'&data_i(0 to width-1)));
	dis_in <= (ac(1 to 17)& io(0)) when dif(0)='1'
		else (dif(2 to width)&io(0));
		
	with ac_sel select
	ac_in <= (ac or data_i) when ior_ac,
		(ac xor data_i) when xor_ac,
		(ac and data_i) when and_ac,
		sum(1 to width) when sum_ac,
		dif(1 to width) when dif_ac,
		(ac(17) & ac(0 to 16)) when rar_ac,
		(io(17) & ac(0 to 16)) when rcr_ac,
		('0' & ac(0 to 16)) when sar_ac,
		(ac(1 to 17)& ac(0)) when ral_ac,
		(ac(1 to 17)& io(0)) when rcl_ac,
		(ac(1 to 17)& '0') when sal_ac,
		dis_in when dis_ac,
		('0'&sum(1 to width-1)) when mus_ac,
		data_i(0 to 5)&ac(6 to width-1) when addr_ac,
		ac(0 to 5)&data_i(6 to width-1) when code_ac,
		data_i(5)&data_i(5)&data_i(5)&data_i(5)&data_i(5)&data_i(5 to width-1) when law_ac,
		"000000"&p when p_ac,
		"000000000000000000" when zero_ac,
		(ac xor "111111111111111111") when cma_ac,
		(ac(0 to width-1)+1) when idx_ac,
		(ac(0 to width-1)-1) when ddx_ac,
		data_i when others;

	with io_sel select
	io_in <= (sum(width)&io(0 to width-2)) when mus_io,
		(ac(width-1)&io(0 to width-2)) when scr_io,
		(io(width-1)&io(0 to width-2)) when rir_io,
		('0'&io(0 to width-2)) when sir_io,
		(io(1 to width-1)&ac(0)) when rcl_io,
		(io(1 to width-1)&io(0)) when ril_io,
		(io(1 to width-1)&'0') when sil_io,
		(io(1 to width-1)&(not(dif(0)))) when dis_io,
		"000000000000000000" when cli_io,
		data_i when others;

	with flag_sel select
	flags <= "000000" when "000",
		"000001" when "001",
		"000010" when "010",
		"000100" when "011",
		"001000" when "100",
		"010000" when "101",
		"100000" when "110",
		"111111" when others;

	with addr_sel select
	addr_o <= a when a_addr,
		p when others;
		
	with ov_sel select
	ov_in <= dif(0) when dif_ov,
		sum(0) when others;
		
	ac_zero <= '1' when(ac = "000000000000000000")
		else '0';
	io_zero <= '1' when(io = "000000000000000000")
		else '0';
	condition <= (not(io(0)))&ov&ac(0)&not(ac(0))&ac_zero;
	
-- sequential assignments, with phase and code
	decode: process(code,a,ac,i,p,io,phase,next_phase,ov)
	begin
		ac_sel<=data_ac; 
		p_sel<=p1_p;
		io_sel<=data_io;
		addr_sel<=p_addr; 
		data_sel<=ac_data; 
		ov_sel<=sum_ov;
		acload<='0'; 
		aload<='0';
		iload<='0';
		pload<='1'; 
		ioload<='0'; 
		ovload<='0';
		mem_write<='0'; 
		mem_read<='1'; 
		io_write<='0'; 
		io_read<='0'; 
		if code < "11010" then	
			if phase = 0 then
				iload <= '1';
				aload <= '1';
				if data_i(5)='1' then
					next_phase <= 1;
				else
					next_phase <= 2;
				end if;
			elsif phase = 1 then
				pload <= '0';
				addr_sel <= a_addr;
				aload <= '1';
				iload <= '0';
				next_phase <= 2;
			else				-- phase=2 or more
				pload <= '0';
				addr_sel <= a_addr;
				acload <= '1';
				case code is
					when add =>
						ac_sel <= sum_ac;
						ov_sel <= sum_ov;
						ovload <= '1';
						next_phase <= 0;
					when sub =>
						ac_sel <= dif_ac;
						ov_sel <= dif_ov;
						ovload <= '1';
						next_phase <= 0;
					when andd =>
						ac_sel <= and_ac;
						next_phase <= 0;
					when iorr =>
						ac_sel <= ior_ac;
						next_phase <= 0;
					when xorr =>
						ac_sel <= xor_ac;
						next_phase <= 0;
					when dis =>
						io_sel <= dis_io;
						ac_sel <= dis_ac;
						ioload <= '1';
						next_phase <= 0;
					when mus =>
						if io(17)='1' then
							ac_sel <= mus_ac;
							io_sel <= mus_io;
						else
							ac_sel <= sar_ac;
							io_sel <= scr_io;
						end if;
						next_phase <= 0;
						ioload <= '1';
					when idx =>
						if phase = 2 then
							addr_sel <= a_addr;
							ac_sel <= data_ac;
							next_phase <= 3;
						else
							addr_sel <= a_addr;
							ac_sel <= idx_ac;
							data_sel <= alu_data;
							mem_write <= '1';
							mem_read <= '0';
							next_phase <= 0;
						end if;
					when isp =>
						if phase = 2 then
							addr_sel <= a_addr;
							ac_sel <= data_ac;
							next_phase <= 3;
						else
							addr_sel <= a_addr;
							ac_sel <= idx_ac;
							data_sel <= alu_data;
							mem_write <= '1';
							mem_read <= '0';
							if ac_in(0)='0' then
								p_sel <= p1_p;
								pload <= '1';
							end if;
							next_phase <= 0;
						end if;
					when ddx =>
						if phase = 2 then
							addr_sel <= a_addr;
							ac_sel <= data_ac;
							next_phase <= 3;
						else
							addr_sel <= a_addr;
							ac_sel <= ddx_ac;
							data_sel <= alu_data;
							mem_write <= '1';
							mem_read <= '0';
							next_phase <= 0;
						end if;
					when dsp =>
						if phase = 2 then
							addr_sel <= a_addr;
							ac_sel <= data_ac;
							next_phase <= 3;
						else
							addr_sel <= a_addr;
							ac_sel <= ddx_ac;
							data_sel <= alu_data;
							mem_write <= '1';
							mem_read <= '0';
							if ac_zero='0' then
								p_sel <= p1_p;
								pload <= '1';
							end if;
							next_phase <= 0;
						end if;
					when lac =>
						ac_sel <= data_ac;
						next_phase <= 0;
					when dac =>
						acload <= '0';
						data_sel <= ac_data;
						mem_write <= '1';
						mem_read <= '0';
						next_phase <= 0;
					when dap =>
						acload <= '0';
						data_sel <= alu_data;
						mem_write <= '1';
						mem_read <= '0';
						next_phase <= 0;
					when dip =>
						acload <= '0';
						data_sel <= alu_data;
						mem_write <= '1';
						mem_read <= '0';
						next_phase <= 0;
					when lio =>
						ioload <= '1';
						acload <= '0';
						io_sel <= data_io;
						next_phase <= 0;
					when dio =>
						acload <= '0';
						data_sel <= io_data;
						mem_write <= '1';
						mem_read <= '0';
						next_phase <= 0;
					when dzm =>
						acload <= '0';
						data_sel <= zero_data;
						mem_write <= '1';
						mem_read <= '0';
						next_phase <= 0;
					when xct =>
						acload <= '0';
						addr_sel <= a_addr;
						pload <= '0';
						next_phase <= 0;
					when jmp =>
						acload <= '0';
						p_sel <= a_p;
						pload <= '1';
						addr_sel <= a_addr;
						next_phase <= 0;
					when jsp =>		-- p must be incremented beforehand
						ac_sel <= p_ac;
						addr_sel <= a_addr;
						pload <= '1';
						p_sel <= a_p;
						next_phase <= 0;
			--		when cal +>		-- same as jda 100
			--		when jda =>		-- dac Y; jsp Y+1
					when sad =>
						addr_sel <= a_addr;
						if (ac /= data_i) then
							p_sel <= p1_p;
							pload <= '1';
						end if;
						next_phase <= 0;
					when sas =>
						addr_sel <= a_addr;
						if (ac = data_i) then
							p_sel <= p1_p;
							pload <= '1';
						end if;
						next_phase <= 0;
					when others =>
						pload <= '0';			-- halt cpu
						next_phase <= 0;
				end case;
			end if;
		else
			case code is
				when law =>
					ac_sel <= law_ac;
					acload <= '1';
					p_sel <= p1_p;
					pload <= '1';
					next_phase <= 0;
				when sft =>
					pload <= '1';
					p_sel <= p1_p;
					next_phase <= 0;
					case data_i(5 to 8) is
						when ral =>
							acload <= '1';
							ac_sel <= ral_ac;
						when ril =>
							ioload <= '1';
							io_sel <= ril_io;
						when rcl =>
							acload <= '1';
							ac_sel <= rcl_ac;
							ioload <= '1';
							io_sel <= rcl_io;
						when sal =>
							acload <= '1';
							ac_sel <= sal_ac;
						when sil =>
							ioload <= '1';
							io_sel <= sil_io;
						when scl =>
							acload <= '1';
							ac_sel <= sal_ac;
							ioload <= '1';
							io_sel <= sil_io;
						when rar =>
							acload <= '1';
							ac_sel <= rar_ac;
						when rir =>
							ioload <= '1';
							io_sel <= rir_io;
						when rcr =>
							acload <= '1';
							ac_sel <= rcr_ac;
							ioload <= '1';
							io_sel <= scr_io;
						when sar =>
							acload <= '1';
							ac_sel <= sar_ac;
						when sir =>
							ioload <= '1';
							io_sel <= sir_io;
						when scr =>
							acload <= '1';
							ac_sel <= sar_ac;
							ioload <= '1';
							io_sel <= scr_io;
						when others =>
							pload <= '0';
					end case;
				when skp =>
					pload <= '1';
					next_phase <= 0;
					if data_i(7 to 11) /= "00000" then
						if (((data_i(7 to 11) and condition) /= "00000") xor (data_i(5)='1')) then
							p_sel <= p2_p;
						end if ;
					--elsif data_1(12 to 14) /= "000" then
						--flag_sel <= data_i(12 to 14);
						--if ((flags and switch_1) /= "000000" then
							--p_sel <= p2_p;
						--end if ;
					--else data_i(15 to 17) /= "000" then
						--flag_sel <= data_i(15 to 17);
						--if ((flags and flag_i) /= "000000" then
							--p_sel <= p2_p;
						--end if ;
					end if;
				when opr =>
					pload <= '1';
					p_sel <= p1_p;
					next_phase <= 0;
					if data_i(6 to width-1)="100000000000" then
						ioload <= '1';
						io_sel <= cli_io;			-- cli
					end if;
					if data_i(6 to width-1)="001000000000" then
						acload <= '1';
						ac_sel <= cma_ac;			-- cma
					end if;
					if data_i(6 to width-1)="000100000000" then
						pload <= '0';				-- hlt
					end if;
					if data_i(6 to width-1)="000010000000" then
						acload <= '1';
						ac_sel <= zero_ac;			-- cla
					end if;
				when iot =>
					if phase = 0 then
						iload <= '1';
						aload <= '1';
						next_phase <= 1;
					else
						addr_sel <= a_addr;
						next_phase <= 0;
						pload <= '0';
						if i(5) = '1' then				-- io write
							data_sel <= ac_data;
							io_write <= '1';
						else
							io_read <= '1';		
							mem_read <= '0';		
							ac_sel <= data_ac;			-- io read
							acload <= '1';
						end if;
					end if;
				when others =>
					pload <= '0';
			end case;
		end if;
	end process decode;

-- finite state machine, processor control unit	
	sync: process(clk,clr)
	begin
		if clr='1' then -- master reset
			phase <= 0;
--			next_phase <= 0;
			a <= (others => '0');
			ac <= (others => '0');
			ov <= '0';
			i <= (others => '0');
			io <= (others => '0');			
			p <= (others => '0');
		elsif (clk'event and clk='1') then
			phase <= next_phase;
			if aload='1' then
				a <= data_i(6 to width-1);
			end if;
			if acload='1' then
				ac <= ac_in;
			end if;
			if ovload='1' then
				ov <= ov_in;
			end if;
			if iload='1' then
				i <= data_i(0 to width-1);
			end if;
			if ioload='1' then
				io <= io_in;
			end if;
			if pload='1' then
				p <= p_in;
			end if;
		end if;
	end process sync;

end behavioral;
